#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time
from array import array

import usb.core
import usb.util
from usb.core import USBError

from canon import CanonError, protocol, commands
from canon.capture import CanonCapture
from canon.storage import CanonStorage

_log = logging.getLogger(__name__)

# Canon
VENDORID = 0x04a9
# PowerShot G3
PRODUCTID = 0x306e

def find(idVendor=VENDORID, idProduct=PRODUCTID):
    """Find a canon camera on some usb bus, possibly.

    Pass in idProduct for your particular model, default values are for a
    PowerShot G3.

    """
    dev = usb.core.find(idVendor=idVendor, idProduct=idProduct)
    if not dev:
        _log.debug("Unable to find a Canon G3 camera attached to this host")
        return None
    _log.info("Found a Canon G3 on bus %s address %s", dev.bus, dev.address)
    return Camera(dev)

class Camera(object):
    """
    Camera objects are the intended API endpoint. Cameras have two
    properties which provide most of the interesting functionality:

     * :attr:`storage` for filesystem operations, and
     * :attr:`capture` for taking pictures.

    Camera objects also expose generic properties of the camera as attributes:
    :attr:`firmware_version`, :attr:`owner`, :attr:`model`,
    :attr:`camera_time`.
    """

    def __init__(self, device):
        """Connect to a :class:`usb.core.Device`.
        """
        self._device = device
        self._usb = protocol.CanonUSB(device)
        self._storage = CanonStorage(self._usb)
        self._capture = CanonCapture(self._usb)
        self._abilities =None
        self._model = None
        self._owner = None
        self._firmware_version = None

    @property
    def ready(self):
        """Check if the camera has been initialized.

        gphoto2 source claims that this command doesn't change the state
        of the camera and can safely be issued without any side effects.

        """
        if not self._device:
            return False
        try:
            commands.IdentifyCameraCmd().execute(self._usb)
            return True
        except (USBError, CanonError):
            return False

    def initialize(self):
        """Bring the camera into a state where it accepts commands.

        This method needs to be called on newly-created instances in order
        to handshake with the camera and make it listen to us.
        There:
        http://www.graphics.cornell.edu/~westin/canon/ch03.html#sec.USBCameraInit

        """
        try:
            cfg = self._device.get_active_configuration()
            _log.debug("Configuration %s already set.", cfg.bConfigurationValue)
        except USBError, e:
            _log.debug("Will attempt to set configuration now, {}".format(e))
            self._device.set_configuration()
            self._device.set_interface_altsetting()

        # Clear endpoint HALTs, not sure if needed, but doesn't hurt
        for ep in (self._usb.ep_in, self._usb.ep_int, self._usb.ep_out):
            try:
                usb.control.clear_feature(self._device, usb.control.ENDPOINT_HALT, ep)
            except USBError, e:
                _log.info("Clearing HALT on {} failed: {}".format(ep, e))

        # while polling, with a gracious timeout, do the dance
        with self._usb.poller_ctx() as p, self._usb.timeout_ctx(2000):
            camstat = self._usb.control_read(0x55, 1).tostring()
            if camstat not in ('A', 'C'):
                raise CanonError('Some kind of init error, camstat: %s', camstat)

            msg = self._usb.control_read(0x01, 0x58)
            if camstat == 'A':
                _log.debug("Camera was already active")
                self._usb.control_read(0x04, 0x50)
                return camstat

            _log.debug("Camera woken up, initializing")

            msg[0:0x40] = array('B', [0]*0x40)
            msg[0] = 0x10
            msg[0x40:] = msg[-0x10:]
            self._usb.control_write(0x11, msg)
            self._usb.bulk_read(0x44)

            started = time.time()
            while len(p.received) < 0x10:
                time.sleep(0.2)
                if time.time() - started > 3.0: # maybe too long
                    # when this happens we're usually ok to proceed ...
                    #raise CanonError("Waited for interrupt in data for too long!")
                    _log.error("Waited for interrupt data for too long!")
                    break

        for _ in range(3):
            try:
                self.identify()
                break
            except (USBError, CanonError), e:
                _log.debug("identify after init fails: {}".format(e))
        if not self.ready:
            raise CanonError("identify_camera failed too many times")

        commands.GenericLockKeysCmd().execute(self._usb)
        self._storage.initialize()
        self._capture.initialize()

    @property
    def storage(self):
        """Access the camera filesystem API.

        See :class:`CanonStorage`.

        """
        return self._storage

    @property
    def capture(self):
        """Access the remote control&capture API.

        See :class:`CanonCapture`.

        """
        return self._capture

    def identify(self):
        """ Return an (model, owner, version) tuple.
        """
        info = commands.IdentifyCameraCmd().execute(self._usb)
        (self._model, self._owner, self._firmware_version) = info
        return info

    @property
    def owner(self):
        """The owner of this camera, writable.
        """
        if not self._owner:
            return self.identify()[1]
        return self._owner

    @owner.setter
    def owner(self, owner):
        commands.SetOwnerCmd(owner).execute(self._usb)
        self.identify()

    @property
    def model(self):
        """Camera model string.
        """
        if not self._model:
            return self.identify()[0]
        return self._model

    @property
    def firmware_version(self):
        """Firmware version as reported by :class:`IdentifyCameraCmd`.
        """
        if not self._firmware_version:
            return self.identify()[2]
        return self._firmware_version

    @property
    def camera_time(self):
        """Camera time as localized unix timestamp, writable.
        """
        return commands.GetTimeCmd().execute(self._usb)

    @camera_time.setter
    def camera_time(self, new):
        # TODO: convert to local tz, accept datetime
        if new is None:
            new = time.time()
        new = int(new)
        commands.SetTimeCmd(new).execute(self._usb)

    @property
    def on_ac(self):
        """True if the camera is not running on battery power.
        """
        return commands.CheckACPowerCmd().execute(self._usb)

    @property
    def abilities(self):
        """Camera "abilities" -- supported image sizes and compressions.
        """
        if not self._abilities:
            return self.get_abilities()
        return self._abilities

    def get_abilities(self):
        """Fetch "abilities" from camera.

        This data isn't used in the protocol, it seems noone figured out a way
        to do so. Data format is explained `here
        <http://www.graphics.cornell.edu/~westin/canon/ch03s25.html>`_.

        """
        self._abilities = commands.GetPicAbilitiesCmd().execute(self._usb)
        return self._abilities

    def cleanup(self):
        """Dispose resources, disable remote capture mode if necessary.
        """
        if not self._device:
            return
        _log.info("Camera {} being cleaned up".format(self))
        usb.util.dispose_resources(self._device)
        self._device = None
        try:
            if self.capture.active:
                self.capture.stop()
        except:
            pass
        self._usb = None
        self._storage = None
        self._capture = None

    def __repr__(self):
        ret = object.__repr__(self)
        if self.ready:
            try:
                return "<{} v{}>".format(self.model, self.firmware_version)
            except:
                return ret
        else:
            return ret

    def __del__(self):
        self.cleanup()



#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Communicate with a Canon camera, old style.

This protocol implementation is heavily based on
http://www.graphics.cornell.edu/~westin/canon/index.html
and gphoto2's source. Sporadic comments here are mostly copied from
gphoto's source and docs.

-- need to put that somewhere --
wValue differs between operations.
wIndex is always 0x00
wLength is simply the length of data.

"""

import time
import threading
import logging
from array import array
from contextlib import contextmanager

import usb.util
from usb.core import USBError

from canon import CanonError
from canon.util import le32toi, hexdump, itole32a

_log = logging.getLogger(__name__)

MAX_CHUNK_SIZE = 0x1400

COMMANDS = []

class CommandMeta(type):
    def __new__(cls, name, bases, attrs):
        super_new = super(CommandMeta, cls).__new__
        parents = [b for b in bases if isinstance(b, CommandMeta)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        new_class = super_new(cls, name, bases, attrs)
        if new_class.is_complete_command():
            COMMANDS.append(new_class)
        return new_class

class Command(object):
    """A USB camera command.

    Subclasses of :class:`Command` are concrete commands to be executed on the
    camera. Instances thereof can run themselves on a :class:`CanonUSB`
    instance and should each implement some sort of response parsing.

    cmd1, cmd2 and cmd3 define the command to be executed.

    All of the following properties need to be set for a command class:

    ``cmd1`` is a command code.

    ``cmd2`` is `0x11` for storage and `0x12` for control commands.

    ``cmd3`` is `0x201` for fixed-response-length commands and
    `0x202` for variable length.

    """
    __metaclass__ = CommandMeta

    cmd1 = None
    cmd2 = None
    cmd3 = None

    MAX_CHUNK_SIZE = 0x1400

    _cmd_serial = 0

    _required_props = ['cmd1', 'cmd2', 'cmd3']

    @classmethod
    def _next_serial(cls):
        cls._cmd_serial += ((cls._cmd_serial % 8)) or 5 # just playin'
        if cls._cmd_serial > 65530:
            cls._cmd_serial = 0
        return cls._cmd_serial | 0x12<<16

    @classmethod
    def is_complete_command(cls):
        for p in cls._required_props:
            if not hasattr(cls, p):
                return False
            if getattr(cls, p) is None:
                return False
        return True

    def __init__(self, payload=None, serial=None):
        if not self.is_complete_command():
            raise AssertionError('{} is incomplete?'.format(self))
        assert ((isinstance(payload, array) and payload.itemsize == 1)
                    or payload is None)

        self._serial = serial
        payload_length = len(payload) if payload else 0
        self._command_header = self._construct_command_header(payload_length)
        self._payload = payload
        self._response_header = None

    @property
    def command_header(self):
        return self._command_header

    @property
    def payload(self):
        return self._payload if self._payload else array('B')

    @property
    def response_header(self):
        return self._response_header

    @response_header.setter
    def response_header(self, data):
        """TODO: check for the same serial and stuff...
        """
        assert isinstance(data, array)
        assert data.itemsize == 1
        assert len(data) == 0x40
        self._response_header = data

    @property
    def response_status(self):
        raise NotImplementedError()

    @property
    def response_length(self):
        raise NotImplementedError()

    @property
    def serial(self):
        """Return the serial id of this command.

        Generate one if not given in the constructor.

        """
        if self._serial is None:
            self._serial = self._next_serial()
        return self._serial

    @property
    def name(self):
        """Simply the class name for convenient access.
        """
        return self.__class__.__name__

    @property
    def first_chunk_size(self):
        """Return the length of the first chunk of data to be read.

        This differs for different commands, but should always be
        at least 0x40.
        """
        raise NotImplementedError()

    @classmethod
    def from_command_packet(data):
        """Return a command instance from a command packet.

        This is used for parsing sniffed USB traffic.

        """
        if not isinstance(data, array):
            data = array('B', data)
        assert len(data) >= 0x40
        request_size = data[0:4]
        cmd1 = data[0x44]
        cmd2 = data[0x47]
        cmd3 = le32toi(data, 4)
        serial = data[0x4c:0x50]

#        cmd_class =
#        return CanonUSBCommand(cmd1, cmd2, cmd3, serial, data[0x50:])

    def _construct_command_header(self, payload_length):
        """Return the 0x50 bytes to send down the control pipe.

        The structure is described here
        http://www.graphics.cornell.edu/~westin/canon/ch03s02.html

        """
        request_size = itole32a(payload_length + 0x10)

        # we dump a 0x50 (80) byte command block
        # the first 0x40 of which are some kind of standard header
        # the next 0x10 seem to be the header for the next layer
        # but it's all the same for us
        packet = array('B', [0] * 0x50)

        # request size is the total transmitted size - the first 0x40 bytes
        packet[0:4] = request_size

        # 0x02 just works, gphoto2 does magic for other camera classes
        packet[0x40] = 0x02

        packet[0x44] = self.cmd1
        # must do this for newer cameras, just a note
        #packet[0x46] = 0x10 if self.cmd3 == 0x201 else 0x20
        packet[0x47] = self.cmd2
        packet[4:8] = itole32a(self.cmd3)
        packet[0x48:0x48+4] = request_size # yes, again

        # this must be matched in the response
        packet[0x4c:0x4c+4] = itole32a(self.serial)

        return packet

    @classmethod
    def next_chunk_size(cls, remaining):
        """Calculate the size of the next chunk to read.

        See
        http://www.graphics.cornell.edu/~westin/canon/ch03s02.html#par.VarXfers

        """
        if remaining > cls.MAX_CHUNK_SIZE:
            return cls.MAX_CHUNK_SIZE
        elif remaining > 0x40:
            return (remaining // 0x40) * 0x40
        else:
            return remaining

    @classmethod
    def chunk_sizes(cls, bytes_to_read):
        """Yield chunk sizes to read.
        """
        while bytes_to_read:
            chunk = cls.next_chunk_size(bytes_to_read)
            bytes_to_read -= chunk
            yield chunk

    def _send(self, usb):
        """Send a command for execution to the camera.

        This method sends the command header and payload down the control
        pipe, reads the response header and returns an iterator over the
        response payload.

        """
        _log.info("--> {0.name:s} (0x{0.cmd1:x}, 0x{0.cmd2:x}, "
                  "0x{0.cmd3:x}), #{1:0}"
                  .format(self, self.serial & 0x0000ffff))

        # control out, then bulk in the first chunk
        usb.control_write(0x10, self.command_header + self.payload)
        data = usb.bulk_read(self.first_chunk_size)

        # store the response header
        self.response_header = data[:0x40]

        # return an iterator over the response data
        return self._reader(usb, data[0x40:])

    def _reader(self, usb, first_chunk):
        raise NotImplementedError()

    def _parse_response(self, data):
        return data

    def execute(self, usb):
        reader = self._send(usb)
        data = array('B')
        for chunk in reader:
            data.extend(chunk)
        return self._parse_response(data)

    def __repr__(self):
        return '<{} 0x{:x} 0x{:x} 0x{:x} at 0x{:x}>'.format(
                    self.name, self.cmd1, self.cmd2, self.cmd3, hash(self))

class GenericCommand(Command):
    @classmethod
    def is_complete_command(cls):
        return True

    def __init__(self, cmd1, cmd2, cmd3, payload=None, serial=None):
        assert cmd1 and cmd2 and cmd3
        self.cmd1 = cmd1
        self.cmd2 = cmd2
        self.cmd3 = cmd3
        super(GenericCommand, self).__init__(payload, serial)


class VariableResponseCommand(Command):
    cmd3 = 0x202
    first_chunk_size = 0x40

    @property
    def response_length(self):
        """Return the response length, excluding the first 0x40 bytes.
        """
        if not self.response_header:
            raise CanonError("_send() this command first.")
        return le32toi(self.response_header, 6)

    def _reader(self, usb, first_chunk):
        _log.debug("variable response says 0x{:x} bytes follow"
                   .format(self.response_length))
        _log.info("<-- {0.name:s} #{1:0} retlen 0x{2:x} "
                  .format(self, self.serial & 0x0000ffff,
                          self.response_length + 0x40))

        # this is normally empty, but let's make sure
        if first_chunk:
            yield first_chunk

        remaining = self.response_length - len(first_chunk)
        for chunk_size in self.chunk_sizes(remaining):
            yield usb.bulk_read(chunk_size)

class FixedResponseCommand(Command):
    cmd3 = 0x201
    resplen = None # the total data length to read, excluding the first 0x40

    _required_props = Command._required_props + ['resplen']

    @property
    def response_length(self):
        """Extract resplen from the response header

        For cmd3=0x201 (fixed response length) commands word at 0x00 is
        response length excluding the first 0x40, as well as the word at
        0x48, the header of the next layer?

        """
        if not self.response_header:
            raise CanonError("_send() this command first.")
        return le32toi(self.response_header, 0)

    @property
    def first_chunk_size(self):
        return self.next_chunk_size(self.resplen + 0x40)

    def _correct_resplen(self, already_got=0):
        if self.response_length != self.resplen:
            _log.warn("BAD response length, correcting 0x{:x} to 0x{:x} "
                      .format(self.resplen, self.response_length))
        return self.response_length - already_got

    def _reader(self, usb, first_chunk):
        remaining = self._correct_resplen(len(first_chunk))

        if len(first_chunk) < 0x0c:
            # need another chunk to get to the response length
            chunk_len = self.next_chunk_size(remaining)
            first_chunk.extend(usb.bulk_read(chunk_len))
            remaining -= chunk_len

        assert len(first_chunk) >= 0x0c

        # word at 0x50 is status byte
        self.status = le32toi(first_chunk, 0x10)
        _log.info("<-- {0.name:s} #{1:0} status: 0x{2:x} "
                  .format(self, self.serial & 0x0000ffff,
                          self.status))

        yield first_chunk

        for chunk_size in self.chunk_sizes(remaining):
            yield usb.bulk_read(chunk_size)


class InterruptPoller(threading.Thread):
    """Poll the interrupt pipe on a CanonUSB.

    This should not be instantiated directly, but via CanonUSB.poller
    """
    def __init__(self, usb, size=None, chunk=0x10, timeout=None):
        threading.Thread.__init__(self)
        self.usb = usb
        self.should_stop = False
        self.size = size
        self.chunk = chunk
        self.received = array('B')
        self.timeout = int(timeout) if timeout is not None else 150
        self.setDaemon(True)

    def run(self):
        errors = 0
        while errors < 10:
            if self.should_stop: return
            try:
                chunk = self.usb.interrupt_read(self.chunk, self.timeout)
                if chunk:
                    self.received.extend(chunk)
                if (self.size is not None
                        and len(self.received) >= self.size):
                    _log.info("poller got 0x{:x} bytes, needed 0x{:x}"
                              ", exiting".format(len(self.received),
                                                 self.size))
                    return
                if self.should_stop:
                    _log.info("poller stop requested, exiting")
                    return
                time.sleep(0.1)
            except (USBError, ) as e:
                if e.errno == 110: # timeout, ignore
                    continue
                if e.errno == 16: # resource busy, bail
                    _log.warn("interrupt pipe busy: {}".format(e))
                    return
                _log.warn("poll: {}".format(e))
                errors += 1
        _log.info("poller got too many errors, exiting")

    def stop(self):
        self.should_stop = True
        self.join()

class CanonUSB(object):
    """USB Link to the camera.
    """
    def __init__(self, device):
        self.max_chunk_size = MAX_CHUNK_SIZE
        self.device = device
        self.device.default_timeout = 500
        self.iface = iface = device[0][0,0]

        # Other models may have different endpoint addresses
        self.ep_in = usb.util.find_descriptor(iface, bEndpointAddress=0x81)
        self.ep_out = usb.util.find_descriptor(iface, bEndpointAddress=0x02)
        self.ep_int = usb.util.find_descriptor(iface, bEndpointAddress=0x83)
        self._cmd_serial = 0
        self._poller = None

    @contextmanager
    def timeout_ctx(self, new):
        old = self.device.default_timeout
        self.device.default_timeout = new
        _log.info("timeout_ctx: {} ms -> {} ms".format(old, new))
        now = time.time()
        try:
            yield
        finally:
            _log.info("timeout_ctx: {} ms <- {} ms; back in {:.3f} ms"
                      .format(old, new, (time.time() - now) * 1000))
            self.device.default_timeout = old

    def start_poller(self, size=None, timeout=None):
        if self._poller and self._poller.isAlive():
            raise CanonError("Poller already started.")
        self._poller = InterruptPoller(self, size, timeout=timeout)
        self._poller.start()

    def stop_poller(self):
        if not self._poller:
            raise CanonError("There's no poller to stop.")
        if self._poller.isAlive():
            self._poller.stop()
        self._poller = None

    @property
    def is_polling(self):
        return bool(self._poller and self._poller.isAlive())

    @property
    def poller(self):
        return self._poller

    @contextmanager
    def poller_ctx(self, size=None, timeout=None):
        if self.is_polling:
            _log.warn("poller_ctx: -> entered while poller was active!")
            yield self._poller
            _log.warn("poller_ctx: <- pretending this isn't real")
        else:
            started = time.time()
            _log.info("poller_ctx: -> ({}, {})".format(size, timeout))
            self.start_poller(size, timeout)
            yield self._poller
            if self.is_polling:
                self.stop_poller()
            _log.info("poller_ctx: <- back in {:.3} ms"
                      .format((time.time() - started) * 1000))

    def control_read(self, wValue, data_length=0, timeout=None):
        """Read from the control pipe.

        ``bRequest`` is 0x4 if length of data is >1, 0x0c otherwise (length >1 ? 0x04 : 0x0C)
        ``bmRequestType`` is 0xC0 during read and 0x40 during write.

        """
        #
        bRequest = 0x04 if data_length > 1 else 0x0c
        _log.info("control_read (req: 0x{:x} wValue: 0x{:x}) reading 0x{:x} bytes"
                   .format(bRequest, wValue, data_length))

        response = self.device.ctrl_transfer(
                                 0xc0, bRequest, wValue=wValue, wIndex=0,
                                 data_or_wLength=data_length, timeout=timeout)
        if len(response) != data_length:
            raise CanonError("incorrect response length form camera")
        _log.debug('\n' + hexdump(response))
        return response

    def control_write(self, wValue, data='', timeout=None):
        # bRequest is 0x4 if length of data is >1, 0x0c otherwise (length >1 ? 0x04 : 0x0C)
        bRequest = 0x04 if len(data) > 1 else 0x0c
        _log.info("control_write (rt: 0x{:x}, req: 0x{:x}, wValue: 0x{:x}) 0x{:x} bytes"
                  .format(0x40, bRequest, wValue, len(data)))
        _log.debug("\n" + hexdump(data))
        # bmRequestType is 0xC0 during read and 0x40 during write.
        i = self.device.ctrl_transfer(0x40, bRequest, wValue=wValue, wIndex=0,
                                      data_or_wLength=data, timeout=timeout)
        if i != len(data):
            raise CanonError("control write was incomplete")
        return i

    def bulk_read(self, size, timeout=None):
        start = time.time()
        data = self.ep_in.read(size, timeout)
        end = time.time()
        data_size = len(data)
        if not data_size == size:
            _log.warn("bulk_read: WRONG SIZE: 0x{:x} bytes instead of 0x{:x}"
                      .format(data_size, size))
            _log.debug('\n' + hexdump(data))
            raise CanonError("unexpected data length ({} instead of {})"
                          .format(len(data), size))
        _log.info("bulk_read got {} (0x{:x}) b in {:.6f} sec"
                  .format(len(data), len(data), end-start))
        _log.debug("\n" + hexdump(data))
        return data

    def interrupt_read(self, size, timeout=100, ignore_timeouts=False):
        try:
            data = self.ep_int.read(size, timeout)
        except USBError, e:
            if ignore_timeouts and e.errno == 110:
                return array('B')
            raise
        if data is not None and len(data):
            dlen = len(data)
            _log.info("interrupt_read: got 0x{:x} bytes".format(dlen))
            _log.debug("\n" + hexdump(data))
            return data
        return array('B')




#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from canon import protocol, commands
from canon.util import extract_string, le32toi, itole32a
from canon.bitfield import BooleanFlag, Bitfield
from array import array
import itertools

_log = logging.getLogger(__name__)

class FSAttributes(Bitfield):

    _size = 0x01

    DOWNLOADED = 0x20
    WRITE_PROTECTED = 0x01
    RECURSE_DIR = 0x80
    NONRECURSE_DIR = 0x10

    UNKNOWN_2 = 0x02
    UNKNOWN_4 = 0x04
    UNKNOWN_8 = 0x08
    UNKNOWN_40 = 0x40

    recurse = BooleanFlag(0, true=RECURSE_DIR, false=NONRECURSE_DIR)
    downloaded = BooleanFlag(0, true=DOWNLOADED)
    protected = BooleanFlag(0, true=WRITE_PROTECTED)

    @property
    def is_dir(self):
        return (self.RECURSE_DIR in self.recurse
                    or self.NONRECURSE_DIR in self.recurse)

class FSEntry(object):
    def __init__(self, name, attributes, size=None, timestamp=None):
        self.name = name
        self.size = size
        self.timestamp = timestamp
        if not isinstance(attributes, FSAttributes):
            attributes = FSAttributes(attributes)
        self.attr = attributes
        self.children = []
        self.parent = None

    @property
    def full_path(self):
        if self.parent is None:
            return self.name
        return self.parent.full_path + '\\' + self.name
    @property
    def entry_size(self):
        return 11 + len(self.name)

    @property
    def type_(self):
        return 'd' if self.attr.is_dir else 'f'

    @property
    def is_dir(self):
        return self.attr.is_dir

    @property
    def is_file(self):
        return not self.attr.is_dir

    def __iter__(self):
        yield self
        for entry in itertools.chain(*self.children):
            yield entry

    def __repr__(self):
        return "<FSEntry {0.type_} '{0.full_path}'>".format(self)

    def __str__(self):
        return self.full_path

class ListDirectoryCmd(commands.VariableResponseCommand):
    cmd1 = 0x0b
    cmd2 = 0x11
    def __init__(self, path=None, recurse=12):
        payload = array('B', [recurse])
        payload.extend(array('B', path))
        payload.extend(array('B', [0x00] * 3))
        super(ListDirectoryCmd, self).__init__(payload)

    def _entry_generator(self, data):
        idx = 0
        while True:
            name = extract_string(data, idx+10)
            if not name:
                raise StopIteration()
            entry = FSEntry(name, attributes=data[idx:idx+1],
                            size=le32toi(data[idx+2:idx+6]),
                            timestamp=le32toi(data[idx+6:idx+10]))
            idx += entry.entry_size
            yield entry

    def _parse_response(self, data):
        entry_generator = self._entry_generator(data)
        root = entry_generator.next()
        current = root
        for entry in entry_generator:
            if entry.name == '..':
                current = current.parent
                continue
            current.children.append(entry)
            entry.parent = current
            if entry.name.startswith('.\\'):
                entry.name = entry.name[2:]
                current = entry
            _log.info(entry)

        return root


class GetFileCmd(commands.VariableResponseCommand):
    cmd1 = 0x01
    cmd2 = 0x11
    def __init__(self, path, target, thumbnail=False):
        payload = array('B', [0x00]*8)
        payload[0] = 0x01 if thumbnail else 0x00
        payload[4:8] = itole32a(protocol.MAX_CHUNK_SIZE)
        payload.extend(array('B', path))
        payload.append(0x00)
        self._target = target
        super(GetFileCmd, self).__init__(payload)
    def execute(self, usb):
        reader = self._send(usb)
        for chunk in reader:
            self._target.write(chunk)


class CanonStorage(object):
    def __init__(self, usb):
        self._usb = usb
        self._drive = None

    def initialize(self, force=False):
        self.get_drive()
        #self.get_disk_info()

    def get_disk_info(self):
        raise NotImplementedError()

    def get_drive(self):
        """Returns the Windows-like camera FS root.
        """
        self._drive = commands.IdentifyFlashDeviceCmd().execute(self._usb)
        return self._drive

    @property
    def drive(self):
        if not self._drive:
            self.get_drive()
        return self._drive

    def ls(self, path=None, recurse=12):
        """Return a class:`FSEntry` for the path or storage root.

        By default this will return the tree starting at ``path`` with large
        enough recursion depth to cover every file on the camera storage.
        """
        path = self._normalize_path(path)
        return ListDirectoryCmd(path, recurse).execute(self._usb)

    def walk(self, path=None):
        """Iterate over camera storage contents, like ``os.walk()``.
        """
        root = self.ls(path)
        queue = [root]
        while len(queue):
            entry = queue.pop(0)
            dirpath = entry.full_path
            dirnames = []
            filenames = []
            for child in entry.children:
                if child.is_dir:
                    dirnames.append(child.name)
                    queue.append(child)
                else:
                    filenames.append(child.name)
            yield (dirpath, dirnames, filenames)

    def get_file(self, path, target, thumbnail=False):
        """Download a file from the camera.

        ``target`` is either a file-like object or the file name to open
        and write to.

        ``thumbnail`` says wheter to get the thumbnail or the whole file.

        """
        if not hasattr(target, 'write'):
            target = open(target, 'wb+')
        path = self._normalize_path(path)
        GetFileCmd(path, target, thumbnail).execute(self._usb)

    def mkdir(self):
        raise NotImplementedError()

    def rmdir(self):
        raise NotImplementedError()

    def put_file(self):
        raise NotImplementedError()

    # ...

    def _normalize_path(self, path):
        if path is None:
            path = ''
        if isinstance(path, FSEntry):
            path = path.full_path
        path = path.replace('/', '\\')
        if not path.startswith(self.drive):
            path = '\\'.join([self.drive, path]).rstrip('\\')
        return path






#
# This file is part of canon-remote
# Copyright (C) 2011 Kiril Zyapkov
#
# canon-remote is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# canon-remote is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with canon-remote.  If not, see <http://www.gnu.org/licenses/>.
#

import struct
import string
from array import array
import math

ARRAY_FORMAT = [None, 'B', '<H', '<I', '<I', '<Q', '<Q', '<Q', '<Q']

def extract_string(data, start=0):
    try:
        end = data[start:].index(0x00)
    except (ValueError, IndexError):
        return None
    return data[start:start+end].tostring()

def le16toi(raw, start=None):
    raw = _normalize_to_string(raw)
    if start is not None:
        raw = raw[start:start+2]
    return struct.unpack('<H', raw)[0]

def le32toi(raw, start=None):
    raw = _normalize_to_string(raw)
    if start is not None:
        raw = raw[start:start+4]
    return struct.unpack('<I', raw)[0]

def itole32a(i):
    return array('B', struct.pack('<I', i))

def _normalize_to_string(raw):
    if isinstance(raw, array):
        raw = raw.tostring()
    elif type(raw) is str:
        pass
    elif type(raw) in (int, long):
        length = math.ceil(raw.bit_length() / 8.0)
        raw = array(ARRAY_FORMAT[length], [raw]).tostring()
    return raw

def chunks(l, n):
    """ Yield successive n-sized chunks from l.
    """
    for i in xrange(0, len(l), n):
        yield l[i:i+n]

def hexdump(data, with_ascii=True, with_offset=True):
    """Return the binary data as nicely printed hexadecimal text.
    """
    if type(data) is str:
        data = array('B', data)
    elif type(data) is unicode:
        data = array('u', data)

    data = enumerate(chunks(data, 0x10)) # 16 bytes per line

    def format_row(idx, row):
        'line of text for the 16 bytes in row'
        line = ''
        if with_offset:
            line = '{:04x}  '.format(idx*0x10)

        halfs = []
        for half in chunks(row, 8): # split the 16 bytes in the middle
            half = ' '.join("{:02x}".format(x) for x in half)
            halfs.append(half)

        line += '  '.join(halfs)

        if not with_ascii:
            return line

        line = line.ljust(57) # adjust to 56 chars
        halfs = []
        for half in chunks(row, 8):
            chars = [(chr(c) if (chr(c) in string.ascii_letters
                                     or chr(c) in string.digits
                                     or chr(c) in string.punctuation
                                     or chr(c) == ' ')
                             else ('.' if c == 0x00 else ';'))
                     for c in half]
            half = ''.join(chars)
            halfs.append(half)
        line += ' '.join(halfs)
        return line

    out = [format_row(row_idx, row) for row_idx, row in data]

    return "\n".join(out)




#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
USB Command structures extracted from gphoto2, see
    libgphoto2/camlibs/canon/usb.c for command structures;
    the containing directory for usage thereof.

TODO: Pythonify this code: commands should be objects with the
      basic pack/unpack logic return length calculation and such
      built in.
"""

import logging
from array import array

from canon.protocol import FixedResponseCommand, VariableResponseCommand
from canon.util import itole32a, le32toi, extract_string, le16toi

_log = logging.getLogger(__name__)

class IdentifyCameraCmd(FixedResponseCommand):
    """Identify camera.

    ``execute()`` -> `(model, owner, version)` strings

    """
    cmd1 = 0x01
    cmd2 = 0x12
    resplen = 0x5c
    def _parse_response(self, data):
        model = extract_string(data, 0x1c)
        owner = extract_string(data, 0x3c)
        version = '.'.join([str(x) for x in data[0x1b:0x17:-1]])
        return model, owner, version

class IdentifyFlashDeviceCmd(VariableResponseCommand):
    """Flash device identification.
    """
    cmd1 = 0x0a
    cmd2 = 0x11
    def _parse_response(self, data):
        return extract_string(data)

class GenericLockKeysCmd(FixedResponseCommand):
    """Lock keys and turn off LCD.
    """
    cmd1 = 0x20
    cmd2 = 0x12
    resplen = 0x14

class SetOwnerCmd(FixedResponseCommand):
    """Change camera owner.
    """
    cmd1 = 0x05
    cmd2 = 0x12
    resplen = 0x14
    def __init__(self, owner):
        payload = array('B', owner + '\x00')
        super(SetOwnerCmd, self).__init__(payload)

class GetOwnerCmd(FixedResponseCommand):
    """Doesn't appear to be supported by the G3.
    """
    cmd1 = 0x05
    cmd2 = 0x12
    resplen = 0x34

class GetTimeCmd(FixedResponseCommand):
    cmd1 = 0x03
    cmd2 = 0x12
    resplen = 0x20
    def _parse_response(self, data):
        return le32toi(data, 0x14)

class SetTimeCmd(FixedResponseCommand):
    cmd1 = 0x04
    cmd2 = 0x12
    resplen = 0x14
    def __init__(self, new_timestamp):
        payload = itole32a(new_timestamp) + array('B', [0] * 8)
        super(SetTimeCmd, self).__init__(payload)

class GetPowerStatusCmd(FixedResponseCommand):
    cmd1 = 0x0a
    cmd2 = 0x12
    resplen = 0x18

class CheckACPowerCmd(GetPowerStatusCmd):
    def _parse_response(self, data):
        return bool((data[0x17] & 0x20) == 0x00)

class GetPicAbilitiesCmd(FixedResponseCommand):
    cmd1 = 0x1f
    cmd2 = 0x12
    resplen = 0x354
    def _parse_response(self, data):
        struct_size = le16toi(data, 0x14)
        model_id = le32toi(data[0x16:0x1a])
        camera_name = extract_string(data, 0x1a)
        num_entries = le32toi(data, 0x3a)
        _log.info("abilities of {} (0x{:x}): 0x{:x} long, n={}"
                 .format(camera_name, model_id, struct_size, num_entries))
        offset = 0x3e
        abilities = []
        for i in xrange(num_entries):
            name = extract_string(data, offset)
            height = le32toi(data, offset + 20)
            width = le32toi(data, offset + 24)
            idx = le32toi(data, offset + 28)
            _log.info(" {:-3} - 0x{:04x} {:20} {}x{}"
                      .format(i, idx, name, width, height))
            offset += 32
            abilities.append((idx, name, height, width))

        return abilities

# Regular camera and storage commands

GET_FILE = {
    'c_idx': 'GET_FILE',
    'description': "Get file",
    'cmd1': 0x01,
    'cmd2': 0x11,
    'cmd3': 0x202,
    'return_length': 0x40 }

MKDIR = {
    'c_idx': 'MKDIR',
    'description': "Make directory",
    'cmd1': 0x05,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

RMDIR = {
    'c_idx': 'RMDIR',
    'description': "Remove directory",
    'cmd1': 0x06,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

DISK_INFO = {
    'c_idx': 'DISK_INFO',
    'description': "Disk info request",
    'cmd1': 0x09,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x5c }

FLASH_DEVICE_IDENT = {
    'c_idx': 'FLASH_DEVICE_IDENT',
    'description': "Flash device ident",
    'cmd1': 0x0a,
    'cmd2': 0x11,
    'cmd3': 0x202,
    'return_length': 0x40 }

DELETE_FILE_2 = {
    'c_idx': 'DELETE_FILE_2',
    'description': "Delete file",
    'cmd1': 0x0a,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

GET_DIR = {
    'c_idx': 'GET_DIR',
    'description': "Get directory entries",
    'cmd1': 0x0b,
    'cmd2': 0x11,
    'cmd3': 0x202,
    'return_length': 0x40 }

DELETE_FILE = {
    'c_idx': 'DELETE_FILE',
    'description': "Delete file",
    'cmd1': 0x0d,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

DISK_INFO_2 = {
    'c_idx': 'DISK_INFO_2',
    'description': "Disk info request (new)",
    'cmd1': 0x0d,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x5c }

SET_ATTR = {
    'c_idx': 'SET_ATTR',
    'description': "Set file attributes",
    'cmd1': 0x0e,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

FLASH_DEVICE_IDENT_2 = {
    'c_idx': 'FLASH_DEVICE_IDENT_2',
    'description': "Flash device ident (new)",
    'cmd1': 0x0e,
    'cmd2': 0x11,
    'cmd3': 0x202,
    'return_length': 0x40 }

SET_FILE_TIME = {
    'c_idx': 'SET_FILE_TIME',
    'description': "Set file time",
    'cmd1': 0x0f,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

IDENTIFY_CAMERA = {
    'c_idx': 'IDENTIFY_CAMERA',
    'description': "Identify camera",
    'cmd1': 0x01,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x9c }

GET_TIME = {
    'c_idx': 'GET_TIME',
    'description': "Get time",
    'cmd1': 0x03,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x60 }

SET_TIME = {
    'c_idx': 'SET_TIME',
    'description': "Set time",
    'cmd1': 0x04,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

CAMERA_CHOWN = {
    'c_idx': 'CAMERA_CHOWN',
    'description': "Change camera owner",
    'cmd1': 0x05,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

GET_OWNER = {
    'c_idx': 'GET_OWNER',
    'description': "Get owner name (new)",
    'cmd1': 0x05,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x74 }

CAMERA_CHOWN_2 = {
    'c_idx': 'CAMERA_CHOWN_2',
    'description': "Change owner (new)",
    'cmd1': 0x06,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

POWER_STATUS = {
    'c_idx': 'POWER_STATUS',
    'description': "Power supply status",
    'cmd1': 0x0a,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x58 }

CONTROL_CAMERA = {
    'c_idx': 'RC_CAMERA',
    'description': "Remote camera control",
    'cmd1': 0x13,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x40 }

POWER_STATUS_2 = {
    'c_idx': 'POWER_STATUS_2',
    'description': "Power supply status (new)",
    'cmd1': 0x13,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x58 }

RETRIEVE_CAPTURE = {
    'c_idx': 'RETRIEVE_CAPTURE',
    'description': "Download a captured image",
    'cmd1': 0x17,
    'cmd2': 0x12,
    'cmd3': 0x202,
    'return_length': 0x40 }

RETRIEVE_PREVIEW = {
    'c_idx': 'RETRIEVE_PREVIEW',
    'description': "Download a captured preview",
    'cmd1': 0x18,
    'cmd2': 0x12,
    'cmd3': 0x202,
    'return_length': 0x40 }

UNKNOWN_FUNCTION = {
    'c_idx': 'UNKNOWN_FUNCTION',
    'description': "Unknown function",
    'cmd1': 0x1a,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x80 }

EOS_LOCK_KEYS = {
    'c_idx': 'EOS_LOCK_KEYS',
    'description': "EOS lock keys",
    'cmd1': 0x1b,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

EOS_UNLOCK_KEYS = {
    'c_idx': 'EOS_UNLOCK_KEYS',
    'description': "EOS unlock keys",
    'cmd1': 0x1c,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

EOS_GET_BODY_ID = {
    'c_idx': 'EOS_GET_BODY_ID',
    'description': "EOS get body ID",
    'cmd1': 0x1d,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x58 }

GET_PIC_ABILITIES = {
    'c_idx': 'GET_PIC_ABILITIES',
    'description': "Get picture abilities",
    'cmd1': 0x1f,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x394 }

GENERIC_LOCK_KEYS = {
    'c_idx': 'GENERIC_LOCK_KEYS',
    'description': "Lock keys and turn off LCD",
    'cmd1': 0x20,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

#20D_UNKNOWN_1 = {
#    'c_idx': '20D_UNKNOWN_1',
#    'description': "Unknown EOS 20D function",
#    'cmd1': 0x21,
#    'cmd2': 0x12,
#    'cmd3': 0x201,
#    'return_length': 0x54 }
#
#20D_UNKNOWN_2 = {
#    'c_idx': '20D_UNKNOWN_2',
#    'description': "Unknown EOS 20D function",
#    'cmd1': 0x22,
#    'cmd2': 0x12,
#    'cmd3': 0x201,
#    'return_length': 0x54 }

EOS_GET_BODY_ID_2 = {
    'c_idx': 'EOS_GET_BODY_ID_2',
    'description': "Get body ID (new)",
    'cmd1': 0x23,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x58 }

GET_PIC_ABILITIES_2 = {
    'c_idx': 'GET_PIC_ABILITIES_2',
    'description': "Get picture abilities (new)",
    'cmd1': 0x24,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x474 }

CONTROL_CAMERA_2 = {
    'c_idx': 'RC_CAMERA_2',
    'description': "Remote camera control (new)",
    'cmd1': 0x25,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x40 }

RETRIEVE_CAPTURE_2 = {
    'c_idx': 'RETRIEVE_CAPTURE_2',
    'description': "Download captured image (new)",
    'cmd1': 0x26,
    'cmd2': 0x12,
    'cmd3': 0x202,
    'return_length': 0x40 }

LOCK_KEYS_2 = {
    'c_idx': 'LOCK_KEYS_2',
    'description': "Lock keys (new)",
    'cmd1': 0x35,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x5c }

UNLOCK_KEYS_2 = {
    'c_idx': 'UNLOCK_KEYS_2',
    'description': "Unlock keys (new)",
    'cmd1': 0x36,
    'cmd2': 0x12,
    'cmd3': 0x201,
    'return_length': 0x54 }

SET_ATTR_2 = {
    'c_idx': 'SET_ATTR_2',
    'description': "Set file attributes (new))",
    'cmd1': 0x07,
    'cmd2': 0x11,
    'cmd3': 0x201,
    'return_length': 0x54 }

def lookup(cmd1, cmd2, cmd3):
    for name, c in globals().iteritems():
        if not name.isupper() or type(c) != dict:
            continue
        if not all(((k in c) for k in ('cmd1', 'cmd2', 'cmd3'))):
            continue
        if c['cmd1'] == cmd1 and c['cmd2'] == cmd2 and c['cmd3'] == cmd3:
            return c['c_idx']

# Remote control sub-commands

RC_INIT = {
    'c_idx': 'RC_INIT',
    'description': "Camera control init",
    'value': 0x00,
    'cmd_len': 0x18,
    'return_length': 0x1c }

RC_SHUTTER_RELEASE = {
    'c_idx': 'RC_SHUTTER_RELEASE',
    'description': "Release shutter",
    'value': 0x04,
    'cmd_len': 0x18,
    'return_length': 0x1c }

RC_SET_PARAMS = {
    'c_idx': 'RC_SET_PARAMS',
    'description': "Set release params",
    'value': 0x07,
    'cmd_len': 0x3c,
    'return_length': 0x1c }

RC_SET_TRANSFER_MODE = {
    'c_idx': 'RC_SET_TRANSFER_MODE',
    'description': "Set transfer mode",
    'value': 0x09,
    'cmd_len': 0x1c,
    'return_length': 0x1c }

RC_GET_PARAMS = {
    'c_idx': 'RC_GET_PARAMS',
    'description': "Get release params",
    'value': 0x0a,
    'cmd_len': 0x18,
    'return_length': 0x4c }

RC_GET_ZOOM_POS = {
    'c_idx': 'RC_GET_ZOOM_POS',
    'description': "Get zoom position",
    'value': 0x0b,
    'cmd_len': 0x18,
    'return_length': 0x20 }

RC_SET_ZOOM_POS = {
    'c_idx': 'RC_SET_ZOOM_POS',
    'description': "Set zoom position",
    'value': 0x0c,
    'cmd_len': 0x1c,
    'return_length': 0x1c }

RC_GET_AVAILABLE_SHOT = {
    'c_idx': 'RC_GET_AVAILABLE_SHOT',
    'description': "Get available shot",
    'value': 0x0d,
    'cmd_len': 0x18,
    'return_length': 0x20 }

RC_GET_CUSTOM_FUNC = {
    'c_idx': 'RC_GET_CUSTOM_FUNC',
    'description': "Get custom func.",
    'value': 0x0f,
    'cmd_len': 0x22,
    'return_length': 0x26 }

RC_GET_EXT_PARAMS_SIZE = {
    'c_idx': 'RC_GET_EXT_PARAMS_SIZE',
    'description': "Get ext. release params size",
    'value': 0x10,
    'cmd_len': 0x1c,
    'return_length': 0x20 }

RC_GET_EXT_PARAMS = {
    'c_idx': 'RC_GET_EXT_PARAMS',
    'description': "Get ext. release params",
    'value': 0x12,
    'cmd_len': 0x1c,
    'return_length': 0x2c }

RC_SET_EXT_PARAMS = {
    'c_idx': 'RC_SET_EXT_PARAMS',
    'description': "Set extended params",
    'value': 0x13,
    'cmd_len': 0x15,
    'return_length': 0x1c }

RC_EXIT = {
    'c_idx': 'RC_EXIT',
    'description': "Exit release control",
    'value': 0x01,
    'cmd_len': 0x18,
    'return_length': 0x1c }

RC_UNKNOWN_1 = {
    'c_idx': 'RC_UNKNOWN_1',
    'description': "Unknown remote subcode",
    'value': 0x1b,
    'cmd_len': 0x08,
    'return_length': 0x5e }

RC_UNKNOWN_2 = {
    'c_idx': 'RC_UNKNOWN_2',
    'description': "Unknown remote subcode",
    'value': 0x1c,
    'cmd_len': 0x00,
    'return_length': 0x00 }

RC_VIEWFINDER_START = {
    'c_idx': 'RC_VIEWFINDER_START',
    'description': "Start viewfinder",
    'value': 0x02,
    'cmd_len': 0x00,
    'return_length': 0x00 }

RC_VIEWFINDER_STOP = {
    'c_idx': 'RC_VIEWFINDER_STOP',
    'description': "Stop viewfinder",
    'value': 0x03,
    'cmd_len': 0x00,
    'return_length': 0x00 }

RC_SET_CUSTOM_FUNC = {
    'c_idx': 'RC_SET_CUSTOM_FUNC',
    'description': "Set custom func.",
    'value': 0x0e,
    'cmd_len': 0x00,
    'return_length': 0x00 }

RC_GET_EXT_PARAMS_VER = {
    'c_idx': 'RC_GET_EXT_PARAMS_VER',
    'description': "Get extended params version",
    'value': 0x11,
    'cmd_len': 0x00,
    'return_length': 0x00 }

RC_SELECT_CAM_OUTPUT = {
    'c_idx': 'RC_SELECT_CAM_OUTPUT',
    'description': "Select camera output",
    'value': 0x14,
    'cmd_len': 0x00,
    'return_length': 0x00 }

def lookup_rc(subcmd):
    for name, c in globals().iteritems():
        if (not name.isupper() or not name.startswith('RC_')
                or type(c) != dict):
            continue
        if not all(((k in c) for k in ('c_idx', 'value', 'cmd_len'))):
            continue
        if c['value'] == subcmd:
            return c['c_idx']

if __name__ == '__main__':
    print COMMANDS



#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Use old Canon cameras with Python.
"""

import logging

version_info = (0, 0, 1, 'dev')

__author__ = u'Kiril Zyapkov'
__contact__ = u'<kiril.zyapkov@gmail.com>'
__version__ = '.'.join(map(str, version_info[:3]))
__copyright__ = u'Copyright (c) 2012 Kiril Zyapkov'
__license__   = u'GPLv3'

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

class CanonError(Exception): pass



#
# This file is part of canon-remote
# Copyright (C) 2011 Kiril Zyapkov
#
# canon-remote is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# canon-remote is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with canon-remote.  If not, see <http://www.gnu.org/licenses/>.
#

from array import array
from canon.util import _normalize_to_string, ARRAY_FORMAT
import struct
import inspect

class _BoundFlag(object):
    """An instance binging a ``Flag`` to a ``Bitfield``.
    """
    def __init__(self, bitfield, flag):
        self._bitfield = bitfield
        self._flag = flag
        for attr_name in ('_fmt', '_fmt_size', '_start', '_end', '_length',
                          'name'):
            setattr(self, attr_name, getattr(flag, attr_name))

    def set_(self, value):
        """Store integer value in this bitfield.
        """
        data = self._pack(value)
        self._store(data)

    def __int__(self):
        return self._unpack(self._extract())

    def __hex__(self):
        """hex() needs a push."""
        return hex(int(self))

    def __iadd__(self, other):
        """Set bits from other in self, self |= other"""
        new = int(self) | int(other)
        self.set_(new)
        return self

    def __isub__(self, other):
        """Clear bits from other in self, self &= ^other """
        new = int(self) & (~int(other))
        self.set_(new)
        return self

    def __contains__(self, other):
        """(v1[, v2 ...]) in self <=> all(lambda x: x in self, (v1[, v2 ...]))
        """
        all_ = 0x00
        try:
            for o in other:
                all_ |= int(o)
        except TypeError: # other is not iterable, unless some int(o) raised,
            all_ = int(other) # but if so this should raise too

        return int(self) == all_

    def _extract(self):
        """return a self._length-long array from bitfield
        """
        return self._bitfield[self._start:self._end]

    def _store(self, data):
        """set the bytes of length self._length in the bitfield
        """
        if not isinstance(data, array):
            data = array('B', data)
        assert len(data) == self._length
        self._bitfield[self._start:self._end] = data

    def _pad(self, data):
        if self._length == self._fmt_size:
            return data
        # must add padding bytes for struct.unpack
        pad = '\x00' * (self._fmt_size - self._length)
        if '<' in self._fmt:
            data = data + pad
        else:
            data = pad + data
        return data

    def _unpad(self, data):
        if self._length == self._fmt_size:
            return data
        # must remove padding bytes for struct.pack
        offset = self._fmt_size - self._length
        if '<' in self._fmt:
            return data[:-offset]
        else:
            return data[offset:]

    def _unpack(self, data):
        """Return the data array as number.
        """
        data = _normalize_to_string(data)
        assert len(data) == self._length
        data = self._pad(data)
        return struct.unpack(self._fmt, data)[0]

    def _pack(self, value):
        """Return value as array('B') of length self._length.
        """
        data = struct.pack(self._fmt, int(value))
        return self._unpad(data)

    def __getattr__(self, name):
        return getattr(self._flag, name)

    def __repr__(self):
        return ("<{} {} 0b{:0"+str(self._length)+"b}>").format(
                   self.name, self._flag.get_value_name(int(self)), int(self))


class Flag(object):
    """A set of bitmasks within a bitfield.

    This may represent combinations of bit states between 8 and 64 bits long,
    instances are descriptors on Bitfield-s.

    """

    _bound_class = _BoundFlag

    def __init__(self, offset, length=None, fmt=None, **choices):
        """
        ``offset``
            in bytes, from the beginning of the bitfield.
        ``length``
            in bytes, between 1 and 8, defaults to 1.
        ``fmt``
            the formatting character for pack/unpack. Only specify if
            something exotic, ``ARRAY_FORMAT`` has sensible defaults.
        ``choices``
            contains a map of human-readable labels of different values.

        """
        self._bound = {}
        self._choices = {}
        self._start = int(offset)
        self._length = 1
        if length is not None:
            assert length > 0 and length <= 8
            self._length = int(length)
        self._end = self._start + self._length
        if fmt is None:
            self._fmt = ARRAY_FORMAT[self._length]
        else:
            assert struct.calcsize(fmt) == \
                    struct.calcsize(ARRAY_FORMAT[self._length])
            self._fmt = fmt

        self._fmt_size = struct.calcsize(self._fmt)
        if choices:
            self._choices = dict([(k.lower(), v) for (k, v) in choices.iteritems()])

    def __getattr__(self, name):
        name = name.lower()
        if name in self._choices:
            return self._choices[name]
        raise AttributeError("{} has no attribute {}".format(self, name))

    def get_value_name(self, value):
        for k, v in self._choices.iteritems():
            if v == value:
                return k
        return '<unknown 0x{:x}>'.format(value)

    @classmethod
    def _get_bound_instance(cls, self, bitfield):
        return cls._bound_class(bitfield, self)

    def _get_bound(self, bitfield):
        h = hash(bitfield)
        if h not in self._bound:
            self._bound[h] = self._get_bound_instance(self, bitfield)
        return self._bound[h]

    def __get__(self, bitfield, owner):
        if bitfield is None:
            return self
        return self._get_bound(bitfield)

    def __set__(self, bitfield, value):
        if bitfield is None:
            return
        b = self._get_bound(bitfield)
        b.set_(value)

class BooleanFlag(Flag):
    _size = 1
    class _bound_class(_BoundFlag):
        def __nonzero__(self):
            if int(self) == self._choices['true']:
                return True
            return False
        def set_(self, value):
            if isinstance(value, bool):
                value = (self._choices['true']
                         if value
                         else self._choices['false'])
            data = self._pack(value)
            self._store(data)

    def __init__(self, offset=0, length=None, fmt=None, true=0x01, false=0x00):
        choices = dict(true=true, on=true, y=true, yes=true,
                       false=false, off=false, n=false, no=false)
        if true ^ false == 0x00:
            raise ValueError("Values for true and false must differ")

        super(BooleanFlag, self).__init__(offset, length, fmt, **choices)
        self._mask = true ^ false

class Bitfield(array):
    """Packs an array('B', ...) as a set of flags

    This can be any length and must be subclassed. Subclasses define
    ``cls._size`` to be the length of the bitfield in bytes and any
    number of instances of ``Flag`` as class attributes.

    The ``Flag`` instances provide the descriptor protocol and
    convenient access to values of various flag within the array.
    The bytes they represent are defined by their ``offset`` and
    ``length`` constructor arguments, ``Flag`` instances can overlap
    within a bitfield, but that's probably to be avoided unless absolutely
    necessary.

    """
    _size = None
    def __new__(cls, data=None):
        if cls._size is None:
            raise RuntimeError("Subclasses of Bitfield should define _size")
        if data is None:
            data = [0] * cls._size
        elif len(data) != cls._size:
            raise RuntimeError("Unexpected data length for {}, got {}"
                               .format(cls, len(data)))
        bf = array.__new__(cls, 'B', data)
        bf.flags = {}
        for flag_name, flag in inspect.getmembers(
                          cls, lambda f: isinstance(f, Flag)):
            flag.name = flag_name
            bf.flags[flag_name] = flag
        return bf

    def __repr__(self):
        bounds = []
        for name in self.flags:
            bounds.append(getattr(self, name))
        return "<{} at 0x{:x} {}>".format(
                      self.__class__.__name__, hash(self),
                      ', '.join(['{}'.format(str(bound))
                                 for bound in bounds]))




#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from distutils.core import setup

import canon

README = os.path.join(os.path.dirname(__file__), 'README.rst')

setup(
    name='canon-remote',
    packages=['canon'],
    version=canon.__version__,
    description='Use old Canon cameras with Python.',
    long_description=open(README).read(),
    author='Kiril Zyapkov',
    author_email='kiril.zyapkov@gmail.com',
    url='http://packages.python.org/canon-remote/',
    license='GPLv3',
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Graphics :: Capture",
        "Topic :: Multimedia :: Graphics :: Capture :: Digital Camera",
        "Topic :: System :: Hardware :: Hardware Drivers",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)


#  This file is part of canon-remote.
#  Copyright (C) 2011-2012 Kiril Zyapkov <kiril.zyapkov@gmail.com>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import time
import logging
from array import array

import usb.core

from canon import commands, CanonError
from canon.bitfield import Bitfield, Flag, BooleanFlag
from canon.util import itole32a
from functools import wraps

_log = logging.getLogger(__name__)

class TransferMode(Bitfield):
    THUMB_TO_PC    = 0x01
    FULL_TO_PC     = 0x02
    THUMB_TO_DRIVE = 0x04
    FULL_TO_DRIVE  = 0x08
    _size = 4
    pc = Flag(0, thumb=THUMB_TO_PC, full=FULL_TO_PC)
    drive =  Flag(0, thumb=THUMB_TO_DRIVE, full=FULL_TO_DRIVE)


class CaptureSettings(Bitfield):
    APERTURE_F1_2 = 0x0d
    APERTURE_F1_4 = 0x10
    APERTURE_F1_6 = 0x13
    APERTURE_F1_8 = 0x15
    APERTURE_F2_0 = 0x18
    APERTURE_F2_2 = 0x1b
    APERTURE_F2_5 = 0x1d
    APERTURE_F2_8 = 0x20
    APERTURE_F3_2 = 0x23
    APERTURE_F3_5 = 0x25
    APERTURE_F4_0 = 0x28
    APERTURE_F4_5 = 0x2b
    APERTURE_F5_0 = 0x2d
    APERTURE_F5_6 = 0x30
    APERTURE_F6_3 = 0x33
    APERTURE_F7_1 = 0x35
    APERTURE_F8 = 0x38
    APERTURE_F9 = 0x3b
    APERTURE_F10 = 0x3d
    APERTURE_F11 = 0x40
    APERTURE_F13 = 0x43
    APERTURE_F14 = 0x45
    APERTURE_F16 = 0x48
    APERTURE_F18 = 0x4b
    APERTURE_F20 = 0x4d
    APERTURE_F22 = 0x50
    APERTURE_F25 = 0x53
    APERTURE_F29 = 0x55
    APERTURE_F32 = 0x58

    SHUTTER_SPEED_BULB = 0x04
    SHUTTER_SPEED_30_SEC = 0x10
    SHUTTER_SPEED_25_SEC = 0x13
    SHUTTER_SPEED_20_SEC = 0x15
    SHUTTER_SPEED_15_SEC = 0x18
    SHUTTER_SPEED_13_SEC = 0x1b
    SHUTTER_SPEED_10_SEC = 0x1d
    SHUTTER_SPEED_8_SEC = 0x20
    SHUTTER_SPEED_6_SEC = 0x23
    SHUTTER_SPEED_5_SEC = 0x25
    SHUTTER_SPEED_4_SEC = 0x28
    SHUTTER_SPEED_3_2_SEC = 0x2b
    SHUTTER_SPEED_2_5_SEC = 0x2d
    SHUTTER_SPEED_2_SEC = 0x30
    SHUTTER_SPEED_1_6_SEC = 0x32
    SHUTTER_SPEED_1_3_SEC = 0x35
    SHUTTER_SPEED_1_SEC = 0x38
    SHUTTER_SPEED_0_8_SEC = 0x3b
    SHUTTER_SPEED_0_6_SEC = 0x3d
    SHUTTER_SPEED_0_5_SEC = 0x40
    SHUTTER_SPEED_0_4_SEC = 0x43
    SHUTTER_SPEED_0_3_SEC = 0x45
    SHUTTER_SPEED_1_4 = 0x48
    SHUTTER_SPEED_1_5 = 0x4b
    SHUTTER_SPEED_1_6 = 0x4d
    SHUTTER_SPEED_1_8 = 0x50
    SHUTTER_SPEED_1_10 = 0x53
    SHUTTER_SPEED_1_13 = 0x55
    SHUTTER_SPEED_1_15 = 0x58
    SHUTTER_SPEED_1_20 = 0x5b
    SHUTTER_SPEED_1_25 = 0x5d
    SHUTTER_SPEED_1_30 = 0x60
    SHUTTER_SPEED_1_40 = 0x63
    SHUTTER_SPEED_1_50 = 0x65
    SHUTTER_SPEED_1_60 = 0x68
    SHUTTER_SPEED_1_80 = 0x6b
    SHUTTER_SPEED_1_100 = 0x6d
    SHUTTER_SPEED_1_125 = 0x70
    SHUTTER_SPEED_1_160 = 0x73
    SHUTTER_SPEED_1_200 = 0x75
    SHUTTER_SPEED_1_250 = 0x78
    SHUTTER_SPEED_1_320 = 0x7b
    SHUTTER_SPEED_1_400 = 0x7d
    SHUTTER_SPEED_1_500 = 0x80
    SHUTTER_SPEED_1_640 = 0x83
    SHUTTER_SPEED_1_800 = 0x85
    SHUTTER_SPEED_1_1000 = 0x88
    SHUTTER_SPEED_1_1250 = 0x8b
    SHUTTER_SPEED_1_1600 = 0x8d
    SHUTTER_SPEED_1_2000 = 0x90
    SHUTTER_SPEED_1_2500 = 0x93
    SHUTTER_SPEED_1_3200 = 0x95
    SHUTTER_SPEED_1_4000 = 0x98
    SHUTTER_SPEED_1_5000 = 0x9a
    SHUTTER_SPEED_1_6400 = 0x9d
    SHUTTER_SPEED_1_8000 = 0xA0

    ISO_50 = 0x40
    ISO_100 = 0x48
    ISO_125 = 0x4b
    ISO_160 = 0x4d
    ISO_200 = 0x50
    ISO_250 = 0x53
    ISO_320 = 0x55
    ISO_400 = 0x58
    ISO_500 = 0x5b
    ISO_640 = 0x5d
    ISO_800 = 0x60
    ISO_1000 = 0x63
    ISO_1250 = 0x65
    ISO_1600 = 0x68
    ISO_3200 = 0x70

    AUTO_FOCUS_ONE_SHOT = 0
    AUTO_FOCUS_AI_SERVO = 1
    AUTO_FOCUS_AI_FOCUS = 2
    MANUAL_FOCUS = 3

    FLASH_MODE_OFF = 0
    FLASH_MODE_ON = 1
    FLASH_MODE_AUTO = 2

    BEEP_OFF = 0x00
    BEEP_ON = 0x01

    EXPOSURE_PLUS_2 = 0x10
    EXPOSURE_PLUS_1_ = 0x0d
    EXPOSURE_PLUS_1_1_2 = 0x0c
    EXPOSURE_PLUS_1_1_3 = 0x0b
    EXPOSURE_PLUS_1 = 0x08
    EXPOSURE_PLUS_0_2_3 = 0x05
    EXPOSURE_PLUS_0_1_2 = 0x04
    EXPOSURE_PLUS_0_1_3 = 0x03
    EXPOSURE_ZERO = 0x00
    EXPOSURE_MINUS_0_1_3 = 0xfd
    EXPOSURE_MINUS_0_1_2 = 0xfc
    EXPOSURE_MINUS_0_2_3 = 0xfb
    EXPOSURE_MINUS_1 = 0xf8
    EXPOSURE_MINUS_1_1_3 = 0xf5
    EXPOSURE_MINUS_1_1_2 = 0xf4
    EXPOSURE_MINUS_1_2_3 = 0xf3
    EXPOSURE_MINUS_2 = 0xf0

    # (compression+raw, formats?, size)
    IMAGE_FORMAT_RAW = (0x04, 0x02, 0x00)
    IMAGE_FORMAT_SMALL_NORMAL_JPEG = (0x02, 0x01, 0x02)
    IMAGE_FORMAT_SMALL_FINE_JPEG = (0x03, 0x01, 0x02)
    IMAGE_FORMAT_MEDIUM_NORMAL_JPEG = (0x02, 0x01, 0x01)
    IMAGE_FORMAT_MEDIUM_FINE_JPEG = (0x03, 0x01, 0x01)
    IMAGE_FORMAT_LARGE_NORMAL_JPEG = (0x02, 0x01, 0x00)
    IMAGE_FORMAT_LARGE_FINE_JPEG = (0x03, 0x01, 0x00)
    IMAGE_FORMAT_RAW_AND_SMALL_NORMAL_JPEG = (0x24, 0x12, 0x20)
    IMAGE_FORMAT_RAW_AND_SMALL_FINE_JPEG = (0x34, 0x12, 0x20)
    IMAGE_FORMAT_RAW_AND_MEDIUM_NORMAL_JPEG = (0x24, 0x12, 0x10)
    IMAGE_FORMAT_RAW_AND_MEDIUM_FINE_JPEG = (0x34, 0x12, 0x10)
    IMAGE_FORMAT_RAW_AND_LARGE_NORMAL_JPEG = (0x24, 0x12, 0x00)
    IMAGE_FORMAT_RAW_AND_LARGE_FINE_JPEG = (0x34, 0x12, 0x00)

    _size = 0x2f

    image_format = Flag(offset=1, length=3)
    flash = Flag(offset=0x06, off=0x00, on=0x01, auto=0x02)
    beep = Flag(offset=0x07, on=0x01, off=0x00)
    macro = BooleanFlag(offset=0x0d, true=0x03, false=0x01)
    focus_mode = Flag(offset=0x12, length=1)
    iso = Flag(offset=0x1a, length=1)
    aperture = Flag(offset=0x1c, length=1,
                    F1_2=0x0d, F1_4=0x10, F1_6=0x13, F1_8=0x15,
                    F2_0=0x18, F2_2=0x1b, F2_5=0x1d, F2_8=0x20, F3_2=0x23,
                    F3_5=0x25, F4_0=0x28, F4_5=0x2b, F5_0=0x2d, F5_6=0x30,
                    F6_3=0x33, F7_1=0x35, F8=0x38, F9=0x3b, F10=0x3d,
                    F11=0x40, F13=0x43, F14=0x45, F16=0x48, F18=0x4b,
                    F20=0x4d, F22=0x50, F25=0x53, F29=0x55, F32=0x58)
    shutter_speed = Flag(offset=0x1e)
    exposure_bias = Flag(offset=0x20)
    shooting_mode = Flag(offset=0x08)

def require_active_capture(func):
    @wraps(func)
    def wrapper(inst, *args, **kw):
        if inst.active:
            return func(inst, *args, **kw)
        raise CanonError("{} only works when capture is active"
                         .format(func.__name__))
    return wrapper

class RemoteControlCommand(commands.FixedResponseCommand):
    """Known subcommands are listed here

    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html

    """
    cmd1 = 0x13
    cmd2 = 0x12
    cmd3 = 0x201
    subcmd = None
    subcmd_resplen = None
    _required_props = (commands.FixedResponseCommand._required_props +
                       ['subcmd'])
    first_chunk_size = 0x40

    def __init__(self, payload=None, serial=None):
        cmd_payload = itole32a(self.subcmd)
        if payload is not None:
            cmd_payload.extend(payload)
        else:
            cmd_payload.extend([0x00] * 4) # at least one zero word follows
        super(RemoteControlCommand, self).__init__(cmd_payload, serial)

    @property
    def resplen(self):
        if self.subcmd_resplen is not None:
            return self.subcmd_resplen
        _log.info("{} with unknown resplen, guessing 0x1c".format(self))
        return 0x1c # this seems to be the minimum

class InitRemoteControlCmd(RemoteControlCommand):
    subcmd = 0x00
    subcmd_resplen = 0x1c

class ExitRemoteControlCmd(RemoteControlCommand):
    subcmd = 0x01
    subcmd_resplen = 0x1c

class GetParamsCmd(RemoteControlCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContGetPars
    """
    subcmd = 0x0a
    subcmd_resplen = 0x4c

class SetParamsCmd(RemoteControlCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContSetPars
    """
    subcmd = 0x07
    subcmd_resplen = 0x1c

class GetCaptureSettingsCmd(GetParamsCmd):
    def __init__(self):
        super(GetCaptureSettingsCmd, self).__init__(itole32a(0x04) +
                                                    itole32a(0x00))
    def _parse_response(self, data):
        # RELEASE_PARAMS_LEN in canon.h
        return CaptureSettings(data[0x1c:0x4b])

class SetCaptureSettingsCmd(SetParamsCmd):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContSetPars
    """
    def __init__(self, settings):
        payload = array('B', itole32a(0x30))
        payload.extend(settings)
        super(SetCaptureSettingsCmd, self).__init__(payload)

class SetTransferModeCmd(RemoteControlCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContSetXferMode
    """
    subcmd = 0x09
    subcmd_resplen = 0x1c
    def __init__(self, transfermode):
        if isinstance(transfermode, array):
            assert transfermode.itemsize == 1
            assert len(transfermode) == 4
        else:
            transfermode = itole32a(int(transfermode))

        # XXX:
        # this one probably has more meaning than just setting
        # transfermode flags ...
        payload = itole32a(0x04) + transfermode
        super(SetTransferModeCmd, self).__init__(payload)

class SetZoomPositionCmd(RemoteControlCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContSetZoom
    """
    subcmd = 0x0c
    subcmd_resplen = 0x1c

class ShutterReleaseCmd(RemoteControlCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s18.html#sec.CamContRelShutter
    """
    subcmd = 0x04
    subcmd_resplen = 0x1c

class DownloadCapturedImageCmd(commands.VariableResponseCommand):
    """
    http://www.graphics.cornell.edu/~westin/canon/ch03s19.html
    """
    cmd1 = 0x17
    cmd2 = 0x12
    def __init__(self, image_index, target, thumbnail=False):
        self.target = target
        payload = array('B', [0] * 4)
        payload.extend(itole32a(0x1400))
        payload.extend(itole32a(0x01 if thumbnail else 0x02))
        payload.extend(image_index)
        super(DownloadCapturedImageCmd, self).__init__(payload)

    def execute(self, usb):
        reader = self._send(usb)
        for chunk in reader:
            self._target.write(chunk)


class CanonCapture(object):
    """Manage taking pictures via USB. The whole point.
    """
    def __init__(self, usb):
        self._usb = usb
        self._settings = None
        self._in_rc = False

    def initialize(self, force=False):
        self.stop()

    def start(self, force=False):
        if self._in_rc and not force:
            _log.info("remote capture already active, force me")
            return

        for _ in range(3):
            try:
                if not self._usb.interrupt_read(0x10, ignore_timeouts=True):
                    break
            except usb.core.USBError, e:
                _log.warn("While trying to flush the interrupt pipe: {}"
                          .format(e))
                break
            time.sleep(0.3)

        # if keys are not locked RC INIT fails, maybe
        commands.GenericLockKeysCmd().execute(self._usb)

        # 5 seconds seem more than enough
        with self._usb.timeout_ctx(5000):
            InitRemoteControlCmd().execute(self._usb)
            self._in_rc = True

        # initially set the transfermode to something known
        self.transfermode = TransferMode.FULL_TO_DRIVE

        # wtf is that?
        SetZoomPositionCmd(array('B', [0x04] + [0x00] * 7))

    def stop(self):
        with self._usb.timeout_ctx(1000):
            ExitRemoteControlCmd().execute(self._usb)
            self._in_rc = False

    @property
    def active(self):
        return self._in_rc

    def get_capture_settings(self):
        self._settings = GetCaptureSettingsCmd().execute(self._usb)
        _log.info("capture settings from camera: {}".format(self._settings))
        return self._settings

    @property
    @require_active_capture
    def settings(self):
        if not self._settings:
            return self.get_capture_settings()
        return self._settings

    @property
    @require_active_capture
    def transfer_mode(self):
        """XXX: Remote Capture 2.5.7 seems to do this in a different manner.
        """
        return GetParamsCmd([0x04] + [0x00] * 7).execute(self._usb)
#        return self._usb.do_command_rc(commands.RC_GET_PARAMS, 0x04, 0x00)

    @transfer_mode.setter
    @require_active_capture
    def transfer_mode(self, flags):
        SetTransferModeCmd(flags).execute(self._usb)

    @property
    @require_active_capture
    def shooting_mode(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def iso_speed(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def aperture(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def shutter_speed(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def white_balance(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def flash(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def image_mode(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def flash_compensation(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def exposure_compensation(self):
        raise NotImplementedError()

    @property
    @require_active_capture
    def macro(self):
        return self.settings.macro

    @macro.setter
    @require_active_capture
    def macro(self, enabled):
        enabled = bool(enabled)
        self.settings.macro = enabled
        self._set_capture_settings(self.settings)
        self.get_capture_settings()

    @require_active_capture
    def __call__(self, store_to=None, store_on_camera=None):
        """Capture an image.

            ``store_to``
                should be a writable file-like object which
                also implements the context protocol.
            ``store_on_camera``
                whether to store the image in the internal camera
                storage or not. If `store_to` is given, this is `False`
                by default, otherwise it's `True`.

            If neither argument was supplied, a capture won't happen.
        """
        tm = TransferMode()
        if store_to and not hasattr(store_to, 'write'):
            store_to = open(store_to, 'wb+')

        if store_to: # assume correct type ...
            tm.pc += TransferMode.FULL_TO_PC
            if store_on_camera == True:
                tm.drive += TransferMode.FULL_TO_DRIVE
            self.transfer_mode = tm
            with store_to:
                irq_data = self._capture_command()
                DownloadCapturedImageCmd(irq_data, store_to)

        elif store_on_camera == False:
            raise CanonError("Don't you need the image stored somewhere?")

        else:
            tm.drive += TransferMode.FULL_TO_DRIVE
            self.transfer_mode = tm
            self._capture_command()

    def _capture_command(self):
        with self._usb.poller_ctx() as p:
            ShutterReleaseCmd().execute(self._usb)
            now = time.time()
            while (len(p.received) < 0x10):
                if time.time() - now > 10:
                    _log.warn("Capture is taking longer than 10 seconds ...")
                    return
                time.sleep(1)
            _log.info("Capture completed")
            return p.received

    def _set_capture_settings(self, settings):
        payload = array('B')
        payload.extend(itole32a(0x30))
        payload.extend(settings)
        SetCaptureSettingsCmd(settings).execute(self._usb)
        self.get_capture_settings()

    capture = __call__
